from functools import wraps


def apply_keyfunc(method):
    """A decorator for dict methods that take a key argument.
    This decorator normalizes the key through the instance's keyfunc function.
    
    """
    @wraps(method)
    def wrapped(self, key, *args):
        key = self.keyfunc(key)
        return method(self, key, *args)
    return wrapped


def lenient(keyfunc=None, catch=TypeError):
    """Decorate a keyfunc function by trying to exectute it and if a
    TypeError (or whatever is passed as the catch argument) is raised,
    catch it and return the key unmodified.
    
    """
    def decorator(fn):
        @wraps(fn)
        def wrapped(*args):
            try:
                return fn(*args)
            except catch:
                # In order for this decorator to work both on methods or plain
                # functions, we assume that the key is the last positional
                # argument passed.
                return args[-1]
        return wrapped

    if keyfunc is None:
        return decorator
    return decorator(keyfunc)



from abc import abstractmethod
from collections import MutableMapping, Mapping
from .decorators import apply_keyfunc
from functools import total_ordering
from itertools import repeat, izip


_SENTINEL = object()


@total_ordering
class AbstractNormalizedDict(MutableMapping):
    """A dictionary where keys are normalized through a given function
    before being inserted in the dict.
    
    All of dict's methods have been implemented so it should be possible to use
    it as a drop-in replacement.
    
    Subclasses should define a keyfunc method that takes one argument
    (a key to be inserted/retrieved/deleted) and return a normalized version of it.
    
    """
    @abstractmethod
    def keyfunc(self, key):
        pass

    def __init__(self, map_or_seq=_SENTINEL, **kwargs):
        """Normalize the keys before inserting them in the internal dictionary.
        The signature is (hopefully) the same as the one for dict.
        
        """
        if map_or_seq is _SENTINEL:
            args = []
        elif isinstance(map_or_seq, Mapping):
            args = [((self.keyfunc(k), v) for k, v in map_or_seq.items())]
        else: # sequence of two-tuples
            args = [((self.keyfunc(k), v) for k, v in map_or_seq)]

        kwargs = {self.keyfunc(k): v for k, v in kwargs.iteritems()}
        self._dict = dict(*args, **kwargs)

    def copy(self):
        return type(self)(self.iteritems())

    @apply_keyfunc
    def __getitem__(self, key):
        return self._dict[key]

    @apply_keyfunc
    def __setitem__(self, key, value):
        self._dict[key] = value

    @apply_keyfunc
    def __delitem__(self, key):
        del self._dict[key]

    @apply_keyfunc
    def has_key(self, key):
        return self._dict.has_key(key)

    def __len__(self):
        return len(self._dict)

    def __iter__(self):
        return iter(self._dict)

    def viewitems(self):
        return self._dict.viewitems()

    def viewkeys(self):
        return self._dict.viewkeys()

    def viewvalues(self):
        return self._dict.viewvalues()

    @classmethod
    def fromkeys(cls, seq, value=None):
        return cls(izip(seq, repeat(value)))

    def __cmp__(self, other):
        return cmp(self._dict, other)

    def __lt__(self, other):
        return self._dict < other



from .base import AbstractNormalizedDict
from .decorators import lenient


class NormalizedDict(AbstractNormalizedDict):
    """A normalized dict where the keyfunc is passed in the constructor.
    
    The keyfunc can also be changed on-the-fly.
    
    """
    def __init__(self, keyfunc, *args, **kwargs):
        self._keyfunc = keyfunc
        super(NormalizedDict, self).__init__(*args, **kwargs)

    def copy(self):
        return type(self)(self.keyfunc, self.iteritems())

    @property
    def keyfunc(self):
        return self._keyfunc
    
    @keyfunc.setter
    def keyfunc(self, new_keyfunc):
        old = self.items()
        self.clear()
        self._keyfunc = new_keyfunc
        self.update(old)
        return new_keyfunc


class CaseInsensitiveDict(AbstractNormalizedDict):
    """A dict whose string keys are case-insensitive."""
    @staticmethod
    @lenient
    def keyfunc(key):
        return str.lower(key)


class TruncatedKeyDict(AbstractNormalizedDict):
    """A dict whose keys are truncated to a given length."""
    def __init__(self, key_length, *args, **kwargs):
        self._key_length = key_length
        super(TruncatedKeyDict, self).__init__(*args, **kwargs)

    def copy(self):
        return type(self)(self.key_length, self.iteritems())

    @property
    def key_length(self):
        return self._key_length

    @lenient
    def keyfunc(self, key):
        return key[:self.key_length]


class ModuloKeyDict(AbstractNormalizedDict):
    """A dict whose keys are normalized modulo a given parameter.
    Note: a side-effect of the implementation is that strings containing a
    formatting parameter will be formatted using the modulo.
    
    """
    def __init__(self, modulo, *args, **kwargs):
        self._modulo = modulo
        super(ModuloKeyDict, self).__init__(*args, **kwargs)

    def copy(self):
        return type(self)(self.modulo, self.iteritems())

    @property
    def modulo(self):
        return self._modulo

    @lenient
    def keyfunc(self, key):
        return key % self.modulo



from distutils.core import setup

with open('README.txt') as f:
    long_description = f.read()

setup(
    name='weirdict',
    version='0.3.0',
    author='Baptiste Mispelon',
    author_email='bmispelon@gmail.com',
    packages=['weirdict'],
    url='https://github.com/bmispelon/weirdict',
    license='LICENSE.txt',
    description='Doing weird things with dicts.',
    long_description=long_description,
)
